/*
 * aTunes 1.13.3
 * Copyright (C) 2006-2009 Alex Aranda, Sylvain Gaudard, Thomas Beckers and contributors
 *
 * See http://www.atunes.org/wiki/index.php?title=Contributing for information about contributors
 *
 * http://www.atunes.org
 * http://sourceforge.net/projects/atunes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

package net.sourceforge.atunes.gui.views.dialogs;

import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.List;
import java.util.concurrent.ExecutionException;

import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.SwingWorker;
import javax.swing.WindowConstants;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeWillExpandListener;
import javax.swing.filechooser.FileSystemView;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.ExpandVetoException;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import net.sourceforge.atunes.gui.GuiUtils;
import net.sourceforge.atunes.gui.views.controls.CustomButton;
import net.sourceforge.atunes.gui.views.controls.CustomModalDialog;
import net.sourceforge.atunes.misc.SystemProperties;
import net.sourceforge.atunes.misc.log.Logger;
import net.sourceforge.atunes.utils.LanguageTool;

import org.jvnet.substance.api.renderers.SubstanceDefaultTreeCellRenderer;

/**
 * The Class MultiFolderSelectionDialog. Allows the selection of the repository
 * folders.
 */
public class MultiFolderSelectionDialog extends CustomModalDialog {

    /**
     * Width of a check box. This is used to calculate if user pressed mouse
     * button inside checkbox
     */
    int checkBoxWidth = new JCheckBox().getPreferredSize().width;

    /**
     * The Class CheckNode.
     */
    private class CheckNode extends DefaultMutableTreeNode {

        /** The Constant serialVersionUID. */
        private static final long serialVersionUID = 3563009061452848710L;

        /** The is selected. */
        private boolean isSelected;

        /** The enabled. */
        private boolean enabled = true;

        /** The dir. */
        private Directory dir;

        /** The icon. */
        private Icon icon;

        /**
         * Instantiates a new check node.
         */
        public CheckNode() {
            this(null, null);
        }

        /**
         * Instantiates a new check node.
         * 
         * @param userObject
         *            the user object
         */
        public CheckNode(Object userObject, Icon icon) {
            this(userObject, true, icon);
        }

        /**
         * Instantiates a new check node.
         * 
         * @param userObject
         *            the user object
         * @param allowsChildren
         *            the allows children
         */
        public CheckNode(Object userObject, boolean allowsChildren, Icon icon) {
            super(userObject, allowsChildren);
            this.icon = icon;
            this.setDir((Directory) userObject);
            if (this.getDir() != null && selectedFolders.contains(this.getDir().getFile())) {
                setSelected(true);
            }
        }

        /**
         * Checks if is enabled.
         * 
         * @return true, if is enabled
         */
        public boolean isEnabled() {
            return enabled;
        }

        /**
         * Checks if is selected.
         * 
         * @return true, if is selected
         */
        public boolean isSelected() {
            return isSelected;
        }

        /**
         * Sets the enabled.
         * 
         * @param enabled
         *            the new enabled
         */
        public void setEnabled(boolean enabled) {
            this.enabled = enabled;
        }

        /**
         * Sets the selected.
         * 
         * @param isSelected
         *            the new selected
         */
        public void setSelected(boolean isSelected) {
            this.isSelected = isSelected;

            if (children != null) {
                Enumeration<?> enume = children.elements();
                while (enume.hasMoreElements()) {
                    Object nextElement = enume.nextElement();
                    if (nextElement instanceof CheckNode) {
                        CheckNode node = (CheckNode) nextElement;
                        node.setSelected(isSelected);
                        node.setEnabled(!isSelected);
                    }
                }
            }
            ((DefaultTreeModel) fileSystemTree.getModel()).nodeChanged(this);
        }

        public Icon getIcon() {
            return icon;
        }

        public void setDir(Directory dir) {
            this.dir = dir;
        }

        public Directory getDir() {
            return dir;
        }

    }

    /**
     * The Class CheckRenderer.
     */
    private class CheckRenderer extends SubstanceDefaultTreeCellRenderer {

        /** The Constant serialVersionUID. */
        private static final long serialVersionUID = 5564069979708271654L;

        /** The check. */
        protected JCheckBox check;

        /** The label. */
        protected JLabel label;

        /**
         * Instantiates a new check renderer.
         */
        public CheckRenderer() {
            setLayout(new FlowLayout());
            check = new JCheckBox();
            check.setOpaque(false);
            add(check);
            add(label = new JLabel());
        }

        // TODO this method should be aware of component orientation
        @Override
        public void doLayout() {
            Dimension d_check = check.getPreferredSize();
            Dimension d_label = label.getPreferredSize();
            int y_check = 0;
            int y_label = 0;
            if (d_check.height < d_label.height) {
                y_check = (d_label.height - d_check.height) / 2;
            } else {
                y_label = (d_check.height - d_label.height) / 2;
            }
            check.setLocation(0, y_check);
            check.setBounds(0, y_check, d_check.width, d_check.height);
            label.setLocation(d_check.width, y_label);

            label.setBounds(d_check.width, y_label, d_label.width + 350, d_label.height);
        }

        @Override
        public Dimension getPreferredSize() {
            Dimension d_check = check.getPreferredSize();
            Dimension d_label = label.getPreferredSize();
            return new Dimension(d_check.width + d_label.width, (d_check.height < d_label.height ? d_label.height : d_check.height));
        }

        @Override
        public Component getTreeCellRendererComponent(JTree tree, Object value, boolean isSelected, boolean expanded, boolean leaf, int row, boolean hasFocus) {
            String stringValue = value.toString();
            setEnabled(tree.isEnabled());
            if (value instanceof CheckNode) {
                check.setSelected(((CheckNode) value).isSelected());
                check.setEnabled(((CheckNode) value).isEnabled());
                label.setFont(tree.getFont());
                label.setText(stringValue);
                if (((CheckNode) value).getUserObject() instanceof Directory) {
                    Directory content = (Directory) ((CheckNode) value).getUserObject();
                    label.setIcon(((CheckNode) value).getIcon());
                    if (isInPathOfSelectedFolders(content.getFile()) || ((CheckNode) value).isSelected()) {
                        label.setFont(label.getFont().deriveFont(Font.BOLD));
                    }
                }
            } else if (value instanceof DefaultMutableTreeNode) {
                check.setEnabled(false);
                check.setSelected(false);
                label.setText(stringValue);
                label.setIcon(null);
                label.setFont(tree.getFont());
            }

            return this;
        }

    }

    /**
     * The Class Directory.
     */
    private static class Directory {

        /** The file. */
        private File file;
        /** The display name. */
        private String displayName;

        /**
         * Instantiates a new directory.
         * 
         * @param file
         *            the file
         * @param displayName
         *            the name that should be displayed
         */
        Directory(File file, String displayName) {
            this.setFile(file);
            this.displayName = displayName;
        }

        @Override
        public String toString() {
            return displayName;
        }

        public void setFile(File file) {
            this.file = file;
        }

        public File getFile() {
            return file;
        }

    }

    /** The Constant serialVersionUID. */
    private static final long serialVersionUID = -1612490779910952274L;

    /** The Constant logger. */
    static final Logger logger = new Logger();

    /** The fs view. */
    static FileSystemView fsView = FileSystemView.getFileSystemView();

    /** The file system tree. */
    JTree fileSystemTree;

    /** The scroll pane. */
    JScrollPane scrollPane;

    /** The ok button. */
    JButton okButton;

    /** The cancel button. */
    private JButton cancelButton;

    /** The text. */
    private JLabel text;

    /** The folders selected. */
    List<File> selectedFolders;

    /** The cancelled. */
    boolean cancelled = true;

    /**
     * Instantiates a new multi folder selection dialog.
     * 
     * @param owner
     *            the owner
     */
    public MultiFolderSelectionDialog(JFrame owner) {
        super(owner, 460, 530, true);
        setContent(getContent());
        setResizable(false);
        GuiUtils.applyComponentOrientation(this);
        enableCloseActionWithEscapeKey();
        setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
    }

    /**
     * The main method.
     * 
     * @param args
     *            the arguments
     */
    public static void main(String[] args) {
        MultiFolderSelectionDialog dialog = new MultiFolderSelectionDialog(null);

        dialog.startDialog(null);

        if (!dialog.isCancelled()) {
            List<File> folders = dialog.getSelectedFolders();
            for (File f : folders) {
                System.out.println(f);
            }
        }
    }

    /**
     * Gets the content.
     * 
     * @return the content
     */
    private JPanel getContent() {
        JPanel panel = new JPanel(null);

        text = new JLabel();

        fileSystemTree = new JTree();
        fileSystemTree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
        scrollPane = new JScrollPane();

        okButton = new CustomButton(null, LanguageTool.getString("OK"));
        okButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                cancelled = false;
                dispose();
            }
        });
        cancelButton = new CustomButton(null, LanguageTool.getString("CANCEL"));
        cancelButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                dispose();
            }
        });

        text.setSize(430, 20);
        text.setLocation(10, 10);
        panel.add(text);

        scrollPane.setSize(430, 410);
        scrollPane.setLocation(10, 40);
        panel.add(scrollPane);

        okButton.setSize(100, 25);
        okButton.setLocation(230, 460);
        panel.add(okButton);

        cancelButton.setSize(100, 25);
        cancelButton.setLocation(340, 460);
        panel.add(cancelButton);

        return panel;
    }

    /**
     * Gets the selected folders.
     * 
     * @return the selected folders
     */
    public List<File> getSelectedFolders() {
        return selectedFolders;
    }

    /**
     * Checks if is cancelled.
     * 
     * @return true, if is cancelled
     */
    public boolean isCancelled() {
        return cancelled;
    }

    /**
     * Checks if is in path of selected folders.
     * 
     * @param dir
     *            the dir
     * 
     * @return true, if is in path of selected folders
     */
    boolean isInPathOfSelectedFolders(File dir) {
        String dirPath = dir.getAbsolutePath().concat(SystemProperties.fileSeparator);
        for (File folder : selectedFolders) {
            if (folder.getAbsolutePath().startsWith(dirPath)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Sets the text.
     * 
     * @param text
     *            the new text
     */
    public void setText(String text) {
        this.text.setText(text);
    }

    /**
     * Sets the tree.
     */
    private void setTree() {

        DefaultTreeModel model = new DefaultTreeModel(new DefaultMutableTreeNode(LanguageTool.getString("PLEASE_WAIT") + "..."));
        fileSystemTree.setModel(model);
        fileSystemTree.setRootVisible(true);
        fileSystemTree.setCellRenderer(new CheckRenderer());
        scrollPane.setViewportView(fileSystemTree);
        scrollPane.setVisible(true);

        new SwingWorker<CheckNode, Void>() {

            @Override
            protected CheckNode doInBackground() throws Exception {
                File[] roots = fsView.getRoots();

                CheckNode root = new CheckNode();

                for (File f : roots) {
                    CheckNode treeNode = new CheckNode(new Directory(f, fsView.getSystemDisplayName(f)), fsView.getSystemIcon(f));
                    root.add(treeNode);
                    File[] files = fsView.getFiles(f, true);
                    Arrays.sort(files);
                    for (File f2 : files) {
                        File[] f2Childs = f2.listFiles();
                        if (f2Childs != null) {
                            boolean hasDirs = hasDirectories(f2Childs);
                            CheckNode treeNode2 = new CheckNode(new Directory(f2, fsView.getSystemDisplayName(f2)), fsView.getSystemIcon(f2));
                            treeNode.add(treeNode2);
                            if (hasDirs) {
                                treeNode2.add(new DefaultMutableTreeNode("Dummy node"));
                            }
                        }
                    }
                }
                return root;
            }

            @Override
            protected void done() {
                try {
                    DefaultTreeModel model = new DefaultTreeModel(get());
                    fileSystemTree.setModel(model);
                    fileSystemTree.setRootVisible(false);
                    fileSystemTree.expandRow(0);
                    fileSystemTree.setSelectionRow(0);

                    fileSystemTree.addMouseListener(new MouseAdapter() {
                        @Override
                        public void mouseClicked(MouseEvent e) {
                            if (e.getButton() == MouseEvent.BUTTON1) {
                                int x = e.getX();
                                int y = e.getY();
                                int row = fileSystemTree.getRowForLocation(x, y);
                                TreePath path = fileSystemTree.getPathForRow(row);

                                // If user pressed button over text area, don't select node
                                if (x > fileSystemTree.getPathBounds(path).x + checkBoxWidth) {
                                    return;
                                }

                                if (path != null) {
                                    //fileSystemTree.expandPath(path);
                                    CheckNode node = (CheckNode) path.getLastPathComponent();

                                    if (node.isEnabled()) {

                                        boolean isSelected = !(node.isSelected());
                                        node.setSelected(isSelected);

                                        if (isSelected) {
                                            // Find if another child folder has been added before
                                            List<File> childFolders = new ArrayList<File>();
                                            for (File f : selectedFolders) {
                                                if (f.getAbsolutePath().startsWith(node.getDir().getFile().getAbsolutePath())) {
                                                    childFolders.add(f);
                                                }
                                            }
                                            for (File f : childFolders) {
                                                selectedFolders.remove(f);
                                            }

                                            selectedFolders.add(node.getDir().getFile());
                                        } else {
                                            selectedFolders.remove(node.getDir().getFile());
                                        }

                                        // I need revalidate if node is root.  but why?
                                        if (row == 0) {
                                            fileSystemTree.revalidate();
                                            fileSystemTree.repaint();
                                        }
                                    }
                                }
                            }
                        }
                    });
                    fileSystemTree.addTreeWillExpandListener(new TreeWillExpandListener() {
                        @Override
                        public void treeWillCollapse(TreeExpansionEvent event) throws ExpandVetoException {
                            // Nothing to do
                        }

                        @Override
                        public void treeWillExpand(TreeExpansionEvent event) throws ExpandVetoException {

                            final CheckNode selectedNode = (CheckNode) event.getPath().getLastPathComponent();
                            fileSystemTree.setSelectionPath(event.getPath());
                            selectedNode.removeAllChildren();

                            selectedNode.add(new DefaultMutableTreeNode(LanguageTool.getString("PLEASE_WAIT") + "..."));

                            new SwingWorker<List<CheckNode>, Void>() {

                                @Override
                                protected List<CheckNode> doInBackground() throws Exception {

                                    List<CheckNode> result = new ArrayList<CheckNode>();

                                    Directory dir = (Directory) selectedNode.getUserObject();
                                    File[] files = fsView.getFiles(dir.getFile(), true);
                                    Arrays.sort(files);
                                    for (File f : files) {
                                        File[] fChilds = f.listFiles();
                                        if (fChilds != null) {
                                            boolean hasDirs = hasDirectories(fChilds);

                                            CheckNode treeNode2 = new CheckNode(new Directory(f, fsView.getSystemDisplayName(f)), fsView.getSystemIcon(f));
                                            result.add(treeNode2);
                                            if (hasDirs) {
                                                treeNode2.add(new DefaultMutableTreeNode("Dummy node"));
                                            }
                                            treeNode2.setSelected(selectedNode.isSelected() || selectedFolders.contains(f));
                                            treeNode2.setEnabled(!selectedNode.isSelected());
                                        }
                                    }
                                    return result;
                                }

                                @Override
                                protected void done() {
                                    selectedNode.removeAllChildren();
                                    try {
                                        List<CheckNode> nodes = get();
                                        for (CheckNode node : nodes) {
                                            selectedNode.add(node);
                                        }
                                        ((DefaultTreeModel) fileSystemTree.getModel()).reload(selectedNode);
                                    } catch (InterruptedException e) {
                                        logger.internalError(e);
                                    } catch (ExecutionException e) {
                                        logger.internalError(e);
                                    }
                                }

                            }.execute();

                        }
                    });

                    fileSystemTree.revalidate();
                    fileSystemTree.repaint();
                } catch (Exception e) {
                    logger.internalError(e);
                } finally {
                    okButton.setEnabled(true);
                    // Show default cursor
                    MultiFolderSelectionDialog.this.setCursor(Cursor.getDefaultCursor());
                }

            }

        }.execute();

        // Show wait cursor
        MultiFolderSelectionDialog.this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
        okButton.setEnabled(false);
    }

    /**
     * Checks if an array of files contains at least one directory.
     * 
     * @param files
     *            The array of files that should be checked
     * @return if at least one directory is contained
     */
    boolean hasDirectories(File[] files) {
        for (File file : files) {
            if (file.isDirectory()) {
                return true;
            }
        }
        return false;
    }

    /**
     * Start dialog.
     * 
     * @param selectedFolders
     *            the selected folders
     */
    public void startDialog(List<File> selectedFolders) {
        this.cancelled = true;
        if (selectedFolders == null) {
            this.selectedFolders = new ArrayList<File>();
        } else {
            this.selectedFolders = selectedFolders;
        }
        setTree();
        setVisible(true);
    }

}
