/*
 * aTunes 1.13.4
 * Copyright (C) 2006-2009 Alex Aranda, Sylvain Gaudard, Thomas Beckers and contributors
 *
 * See http://www.atunes.org/wiki/index.php?title=Contributing for information about contributors
 *
 * http://www.atunes.org
 * http://sourceforge.net/projects/atunes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

package net.sourceforge.atunes.gui.views.dialogs.fullScreen;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsDevice;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.util.concurrent.ExecutionException;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.SwingWorker;
import javax.swing.filechooser.FileFilter;

import net.sourceforge.atunes.Constants;
import net.sourceforge.atunes.gui.Fonts;
import net.sourceforge.atunes.gui.GuiUtils;
import net.sourceforge.atunes.gui.images.ImageLoader;
import net.sourceforge.atunes.gui.views.controls.Cover3D;
import net.sourceforge.atunes.gui.views.controls.CustomWindow;
import net.sourceforge.atunes.gui.views.controls.PopUpButton;
import net.sourceforge.atunes.gui.views.controls.playerControls.NextButton;
import net.sourceforge.atunes.gui.views.controls.playerControls.PlayPauseButton;
import net.sourceforge.atunes.gui.views.controls.playerControls.PreviousButton;
import net.sourceforge.atunes.gui.views.controls.playerControls.StopButton;
import net.sourceforge.atunes.gui.views.panels.PlayerControlsPanel;
import net.sourceforge.atunes.kernel.modules.context.ContextHandler;
import net.sourceforge.atunes.kernel.modules.player.PlayerHandler;
import net.sourceforge.atunes.kernel.modules.podcast.PodcastFeedEntry;
import net.sourceforge.atunes.kernel.modules.radio.Radio;
import net.sourceforge.atunes.kernel.modules.repository.audio.AudioFile;
import net.sourceforge.atunes.kernel.modules.state.ApplicationState;
import net.sourceforge.atunes.kernel.modules.visual.VisualHandler;
import net.sourceforge.atunes.misc.SystemProperties;
import net.sourceforge.atunes.misc.log.LogCategories;
import net.sourceforge.atunes.misc.log.Logger;
import net.sourceforge.atunes.model.AudioObject;
import net.sourceforge.atunes.utils.AudioFilePictureUtils;
import net.sourceforge.atunes.utils.ImageUtils;
import net.sourceforge.atunes.utils.LanguageTool;
import net.sourceforge.atunes.utils.StringUtils;
import net.sourceforge.atunes.utils.TimeUtils;

/**
 * The Class FullScreenWindow.
 */
public class FullScreenWindow extends CustomWindow {

    /** The Constant serialVersionUID. */
    private static final long serialVersionUID = 3422799994808333945L;

    static Logger logger = new Logger();

    /** The cover. */
    private Cover3D cover;

    /** The text label. */
    private JLabel textLabel;

    /** The progress bar. */
    JSlider progressBar;

    /** The time. */
    private JLabel time;

    /** The remaining time. */
    private JLabel remainingTime;

    /** The options. */
    private PopUpButton options;

    /** The exit full screen. */
    private JMenuItem exitFullScreen;

    /** The select background. */
    private JMenuItem selectBackground;

    /** The remove background. */
    private JMenuItem removeBackground;

    /** The previous button. */
    private PreviousButton previousButton;

    /** The play button. */
    private PlayPauseButton playButton;

    /** The stop button. */
    private StopButton stopButton;

    /** The next button. */
    private NextButton nextButton;

    /** The playing. */
    private boolean playing;

    /** The current audio object. */
    private AudioObject currentAudioObject;

    /** The background. */
    ImageIcon background;

    /** The key adapter. */
    private KeyAdapter keyAdapter = new KeyAdapter() {
        @Override
        public void keyPressed(KeyEvent e) {
            if (e.getKeyCode() == KeyEvent.VK_F11) {
                setVisible(false);
            }
        }
    };

    /**
     * Instantiates a new full screen dialog.
     * 
     * @param owner
     *            the owner
     */
    public FullScreenWindow(JFrame owner) {
        super(owner, 0, 0);
        setLocation(0, 0);
        setAlwaysOnTop(true);
        setContent();
        addKeyListener(keyAdapter);
        File backgroundFile = null;
        if (ApplicationState.getInstance().getFullScreenBackground() != null) {
            backgroundFile = new File(ApplicationState.getInstance().getFullScreenBackground());
            if (!backgroundFile.exists()) {
                backgroundFile = null;
            }
        }
        if (backgroundFile == null) {
            background = null;
        } else {
            setBackground(backgroundFile);
        }
        GuiUtils.applyComponentOrientation(this);
    }

    /**
     * Gets the inset for progress bar.
     * 
     * @return the inset for progress bar
     */
    private int getInsetForProgressBar() {
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        return (int) (screenSize.width * 0.25);
    }

    /**
     * Gets the remaining time.
     * 
     * @return the remainingTime
     */
    public JLabel getRemainingTime() {
        return remainingTime;
    }

    /**
     * Gets the time.
     * 
     * @return the time
     */
    public JLabel getTime() {
        return time;
    }

    /**
     * Checks if is playing.
     * 
     * @return true, if is playing
     */
    public boolean isPlaying() {
        return playing;
    }

    /**
     * Paint.
     * 
     * @param audioObject
     *            the song
     * @param updateCover
     *            the update cover
     */
    private void paint(final AudioObject audioObject, boolean updateCover) {
        if (audioObject == null) {
            return;
        }
        // If it's a radio, hide progress bar
        boolean isRadio = (audioObject instanceof Radio);
        boolean isPodcastEntry = (audioObject instanceof PodcastFeedEntry);
        progressBar.setVisible(true);
        progressBar.setEnabled(!isRadio && !isPodcastEntry);
        time.setVisible(true);
        remainingTime.setVisible(true);

        if (!isRadio && !isPodcastEntry) {
            textLabel.setText(StringUtils.getString(audioObject.getArtist(), " - ", audioObject.getTitleOrFileName(), " (", TimeUtils.seconds2String(audioObject.getDuration()),
                    ")"));

            if (updateCover) {
                // When changing song, set no image, until update
                cover.setImage(ImageLoader.NO_COVER);

                // Fetch cover
                new SwingWorker<Image, Void>() {
                    @Override
                    protected Image doInBackground() throws Exception {
                        ImageIcon[] pictures = AudioFilePictureUtils.getPicturesForFile((AudioFile) audioObject, -1, -1);
                        if (pictures == null || pictures.length == 0) {
                            return ContextHandler.getInstance().getAlbumCover((AudioFile) audioObject);
                        } else {
                            return pictures[0].getImage();
                        }
                    }

                    @Override
                    protected void done() {
                        Image image;
                        try {
                            image = get();
                            if (image != null) {
                                setPicture((AudioFile) audioObject, image);
                            }
                        } catch (InterruptedException e) {
                            logger.error(LogCategories.IMAGE, e);
                        } catch (ExecutionException e) {
                            logger.error(LogCategories.IMAGE, e);
                        }
                    }
                }.execute();
            }
        } else {
            if (isRadio) {
                textLabel.setText(((Radio) audioObject).getName());
                cover.setImage(ImageLoader.RADIO_BIG);
            } else if (isPodcastEntry) {
                textLabel.setText(((PodcastFeedEntry) audioObject).getName());
                cover.setImage(ImageLoader.RSS_BIG);
            }
        }
    }

    /**
     * Sets the audio object.
     * 
     * @param audioObject
     *            the new audio object
     */
    public void setAudioObject(AudioObject audioObject) {
        if (audioObject == null) {
            textLabel.setText("");
            return;
        }

        if (audioObject == currentAudioObject) {
            return;
        }

        boolean updateCover = currentAudioObject == null || !audioObject.getAlbum().equals(currentAudioObject.getAlbum());

        currentAudioObject = audioObject;

        if (!isVisible()) {
            return;
        }

        paint(audioObject, updateCover);
    }

    /**
     * Sets the background.
     * 
     * @param file
     *            the new background
     */
    void setBackground(File file) {
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        try {
            background = ImageUtils.scaleImageBicubic(ImageIO.read(file), screenSize.width, screenSize.height);
            ApplicationState.getInstance().setFullScreenBackground(file.getAbsolutePath());
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Sets the content.
     */
    private void setContent() {
        JPanel panel = new JPanel(new BorderLayout()) {
            private static final long serialVersionUID = 109708757849271173L;

            @Override
            public void paintComponent(Graphics g) {
                super.paintComponent(g);
                if (background == null) {
                    Graphics2D g2d = (Graphics2D) g;
                    g2d.setPaint(new GradientPaint(0, this.getHeight() / 2f, Color.black, 0, this.getHeight(), Color.LIGHT_GRAY));
                    g2d.fillRect(0, 0, this.getWidth(), this.getHeight());
                } else {
                    Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
                    g.drawImage(ImageUtils.resize(background, screenSize.width, screenSize.height).getImage(), 0, 0, this);
                }
            }
        };
        panel.setBackground(Color.black);
        add(panel);

        options = new PopUpButton(LanguageTool.getString("OPTIONS"), PopUpButton.TOP_RIGHT);
        options.addKeyListener(keyAdapter);
        selectBackground = new JMenuItem(LanguageTool.getString("SELECT_BACKGROUND"));

        selectBackground.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                setVisible(false);
                JFileChooser fileChooser = new JFileChooser();
                fileChooser.setFileFilter(new FileFilter() {
                    @Override
                    public boolean accept(File pathname) {
                        if (pathname.isDirectory()) {
                            return true;
                        }
                        String fileName = pathname.getName().toUpperCase();
                        return fileName.endsWith("JPG") || fileName.endsWith("JPEG") || fileName.endsWith("PNG");
                    }

                    @Override
                    public String getDescription() {
                        return LanguageTool.getString("IMAGES");
                    }
                });
                if (fileChooser.showOpenDialog(FullScreenWindow.this) == JFileChooser.APPROVE_OPTION) {
                    File selectedBackground = fileChooser.getSelectedFile();
                    setBackground(selectedBackground);
                    FullScreenWindow.this.invalidate();
                    FullScreenWindow.this.repaint();
                }
            }
        });

        removeBackground = new JMenuItem(LanguageTool.getString("REMOVE_BACKGROUND"));
        removeBackground.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                background = null;
                ApplicationState.getInstance().setFullScreenBackground(null);
                FullScreenWindow.this.invalidate();
                FullScreenWindow.this.repaint();
            }
        });
        exitFullScreen = new JMenuItem(LanguageTool.getString("CLOSE"));
        exitFullScreen.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                setVisible(false);
            }
        });

        options.add(selectBackground);
        options.add(removeBackground);
        options.add(exitFullScreen);

        previousButton = new PreviousButton(PlayerControlsPanel.PREVIOUS_NEXT_BUTTONS_SIZE);
        playButton = new PlayPauseButton(PlayerControlsPanel.PLAY_BUTTON_SIZE);
        stopButton = new StopButton(PlayerControlsPanel.STOP_MUTE_BUTTONS_SIZE);
        nextButton = new NextButton(PlayerControlsPanel.PREVIOUS_NEXT_BUTTONS_SIZE);

        previousButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                PlayerHandler.getInstance().playPreviousAudioObject();
            }
        });
        playButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                PlayerHandler.getInstance().playCurrentAudioObject(true);
            }
        });
        stopButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                PlayerHandler.getInstance().stopCurrentAudioObject(true);
            }
        });
        nextButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                PlayerHandler.getInstance().playNextAudioObject(false);

            }
        });

        cover = new Cover3D();

        textLabel = new JLabel();
        textLabel.setFont(Fonts.CONTEXT_INFORMATION_BIG_FONT);

        time = new JLabel("0:00");
        time.setFont(new Font(Font.SANS_SERIF, Font.PLAIN, 15));

        remainingTime = new JLabel("0:00");
        remainingTime.setFont(new Font(Font.SANS_SERIF, Font.PLAIN, 15));

        progressBar = new JSlider();
        progressBar.setOpaque(false);
        progressBar.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseReleased(MouseEvent e) {
                if (progressBar.isEnabled()) {
                    int widthClicked = e.getPoint().x;
                    int widthOfProgressBar = progressBar.getSize().width;
                    double perCent = (double) widthClicked / (double) widthOfProgressBar;
                    PlayerHandler.getInstance().seekCurrentAudioObject(perCent);
                }
            }
        });
        progressBar.addKeyListener(keyAdapter);
        progressBar.setToolTipText(LanguageTool.getString("CLICK_TO_SEEK"));
        progressBar.setMinimum(0);
        progressBar.setValue(0);

        JPanel controlsPanel = new JPanel(new GridBagLayout());
        controlsPanel.setOpaque(false);

        GridBagConstraints c = new GridBagConstraints();
        c.gridx = 0;
        c.gridy = 0;
        c.gridwidth = 3;
        controlsPanel.add(textLabel, c);
        int inset = getInsetForProgressBar();

        c.gridy = 1;
        c.insets = new Insets(50, inset, 0, 10);
        c.gridwidth = 1;
        controlsPanel.add(time, c);
        c.gridx = 1;
        c.insets = new Insets(50, 0, 0, 0);
        c.fill = GridBagConstraints.HORIZONTAL;
        c.weightx = 1;
        controlsPanel.add(progressBar, c);
        c.gridx = 2;
        c.insets = new Insets(50, 10, 0, inset);
        c.weightx = 0;
        controlsPanel.add(remainingTime, c);

        // TODO: Add volume controls to full screen
        JPanel buttonsPanel = PlayerControlsPanel.getPanelWithPlayerControls(stopButton, previousButton, playButton, nextButton, null, null, null);

        c.gridx = 0;
        c.gridwidth = 3;
        c.gridy = 2;
        c.insets = new Insets(20, 0, 5, 0);
        controlsPanel.add(buttonsPanel, c);
        c.gridy = 3;
        c.weightx = 0;
        c.fill = GridBagConstraints.NONE;
        c.insets = new Insets(5, 0, 20, 0);
        controlsPanel.add(options, c);

        panel.add(cover, BorderLayout.CENTER);

        panel.add(controlsPanel, BorderLayout.SOUTH);

    }

    /**
     * Sets the full screen.
     * 
     * @param fullscreen
     *            if fullscreen mode should be set
     */
    private void setFullScreen(boolean fullscreen) {
        if (SystemProperties.OS.isWindows() && !SystemProperties.OS.isVista()) {
            if (fullscreen) {
                setSize(GuiUtils.getDeviceWidth(), GuiUtils.getDeviceHeight());
            }
        } else {
            // Get in which screen is application and set full screen in that screen
            GraphicsDevice graphicsDevice = GuiUtils.getGraphicsDeviceForLocation(VisualHandler.getInstance().getFrame().getLocation());
            graphicsDevice.setFullScreenWindow(fullscreen ? this : null);
        }
    }

    /**
     * Sets the max duration.
     * 
     * @param maxDuration
     *            the new max duration
     */
    public void setAudioObjectLength(long maxDuration) {
        progressBar.setMaximum((int) maxDuration);
    }

    /**
     * Sets the picture.
     * 
     * @param file
     *            the file
     * @param picture
     *            the picture
     */
    void setPicture(AudioFile file, Image picture) {
        // If is not current file, discard
        if (file == currentAudioObject) {
            if (picture == null) {
                cover.setImage(ImageLoader.NO_COVER);
            } else {
                ImageIcon imageScaled = ImageUtils.scaleImageBicubic(picture, Constants.FULL_SCREEN_COVER, Constants.FULL_SCREEN_COVER);
                cover.setImage(imageScaled);
            }
        }
    }

    /**
     * Sets the playing.
     * 
     * @param playing
     *            the new playing
     */
    public void setPlaying(boolean playing) {
        this.playing = playing;
        playButton.setPlaying(playing);
    }

    /**
     * Sets the time.
     * 
     * @param time
     *            the time
     * @param totalTime
     *            the total time
     */
    public void setCurrentAudioObjectPlayedTime(long time, long totalTime) {
        long remainingTime1 = totalTime - time;
        if (time == 0) {
            this.remainingTime.setText(TimeUtils.milliseconds2String(time));
        } else {
            this.remainingTime.setText(remainingTime1 > 0 ? StringUtils.getString("- ", TimeUtils.milliseconds2String(remainingTime1)) : "-");
        }

        this.time.setText(TimeUtils.milliseconds2String(time));
        progressBar.setValue((int) time);
    }

    @Override
    public void setVisible(boolean visible) {
        super.setVisible(visible);
        if (visible) {
            paint(currentAudioObject, true);
        }
        setFullScreen(visible);
    }

}
