/*
 * aTunes 1.13.4
 * Copyright (C) 2006-2009 Alex Aranda, Sylvain Gaudard, Thomas Beckers and contributors
 *
 * See http://www.atunes.org/wiki/index.php?title=Contributing for information about contributors
 *
 * http://www.atunes.org
 * http://sourceforge.net/projects/atunes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

package net.sourceforge.atunes.kernel.controllers.osd;

import java.awt.Image;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseListener;

import javax.swing.ImageIcon;
import javax.swing.SwingConstants;
import javax.swing.Timer;

import net.sourceforge.atunes.gui.GuiUtils;
import net.sourceforge.atunes.gui.images.ImageLoader;
import net.sourceforge.atunes.gui.views.dialogs.OSDDialog;
import net.sourceforge.atunes.kernel.controllers.model.WindowController;
import net.sourceforge.atunes.kernel.modules.context.ContextHandler;
import net.sourceforge.atunes.kernel.modules.podcast.PodcastFeedEntry;
import net.sourceforge.atunes.kernel.modules.radio.Radio;
import net.sourceforge.atunes.kernel.modules.repository.audio.AudioFile;
import net.sourceforge.atunes.kernel.modules.state.ApplicationState;
import net.sourceforge.atunes.kernel.modules.visual.VisualHandler;
import net.sourceforge.atunes.misc.SystemProperties;
import net.sourceforge.atunes.model.AudioObject;
import net.sourceforge.atunes.utils.AudioFilePictureUtils;
import net.sourceforge.atunes.utils.StringUtils;
import net.sourceforge.atunes.utils.TimeUtils;

/**
 * The Class OSDDialogController.
 */
public class OSDDialogController extends WindowController<OSDDialog> {

    /**
     * The Enum Direction.
     */
    private enum Direction {

        /** The UP. */
        UP,

        /** The DOWN. */
        DOWN
    }

    /** The Constant STEP. */
    private static final int STEP = 4;

    /** The timer. */
    Timer timer;

    /** The location to stay. */
    Point locationToStay;

    /** The hidden location. */
    Point hiddenLocation;

    /**
     * Instantiates a new oSD dialog controller.
     * 
     * @param dialogControlled
     *            the dialog controlled
     */
    public OSDDialogController(OSDDialog dialogControlled) {
        super(dialogControlled);
        addBindings();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * net.sourceforge.atunes.kernel.controllers.model.Controller#addBindings()
     */
    @Override
    protected void addBindings() {
        MouseListener listener = new OSDDialogMouseListener(getWindowControlled(), this);
        getWindowControlled().addMouseListener(listener);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * net.sourceforge.atunes.kernel.controllers.model.Controller#addStateBindings
     * ()
     */
    @Override
    protected void addStateBindings() {
        // Nothing to do
    }

    /**
     * Gets the timer.
     * 
     * @return the timer
     */
    private Timer getTimer() {
        if (SystemProperties.isJava6Update10OrLater) {
            return new Timer(10, new ActionListener() {

                private static final int FADE = 50;
                private Direction direction = Direction.UP;
                private int show = ApplicationState.getInstance().getOsdDuration() * 100;
                private int wait;
                private int n;

                @Override
                public void actionPerformed(ActionEvent e) {
                    getWindowControlled().setLocation(locationToStay);

                    if (n < FADE && direction == Direction.UP) {
                        GuiUtils.setWindowOpacity(getWindowControlled(), n / (float) FADE);
                        n++;
                        if (n == FADE) {
                            direction = Direction.DOWN;
                        }
                    } else {
                        if (wait < show) {
                            wait++;
                        } else {
                            if (n > 0 && direction == Direction.DOWN) {
                                GuiUtils.setWindowOpacity(getWindowControlled(), n / (float) FADE);
                                n--;
                            } else {
                                timer.stop();
                                getWindowControlled().dispose();
                            }
                        }
                    }
                }
            });
        } else {
            return new Timer(10, new ActionListener() {

                private boolean showAnimation = ApplicationState.getInstance().isAnimateOSD();
                private Direction direction = Direction.UP;
                private int show = ApplicationState.getInstance().getOsdDuration() * 100;
                private int wait;
                private int position;

                @Override
                public void actionPerformed(ActionEvent e) {

                    if (showAnimation && direction == Direction.UP && getWindowControlled().getLocation().y > locationToStay.y) {
                        // Move up
                        getWindowControlled().setLocation(locationToStay.x, getWindowControlled().getLocation().y - STEP);
                    } else {
                        direction = Direction.DOWN;
                        // Show
                        if (++wait > show) {
                            if (showAnimation && direction == Direction.DOWN && getWindowControlled().getLocation().y < hiddenLocation.y) {
                                // Workaround for Gnome	
                                if (position == getWindowControlled().getLocation().y) {
                                    showAnimation = false;
                                }
                                position = getWindowControlled().getLocation().y;
                                // Move down
                                getWindowControlled().setLocation(locationToStay.x, getWindowControlled().getLocation().y + STEP);
                            } else {
                                // Finished
                                direction = Direction.UP;
                                wait = 0;
                                getWindowControlled().dispose();
                                timer.stop();
                            }
                        }
                    }

                }
            });
        }

    }

    @Override
    protected OSDDialog getWindowControlled() {
        return super.getWindowControlled();
    }

    /**
     * Stop animation and optionally (or if app is not using fade animation)
     * hide dialog
     * 
     * @param hide
     */
    public void stopAnimation(boolean hide) {
        if (timer != null && timer.isRunning()) {
            if (hide || !SystemProperties.isJava6Update10OrLater) {
                getWindowControlled().dispose();
            }
            timer.stop();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * net.sourceforge.atunes.kernel.controllers.model.Controller#notifyReload()
     */
    @Override
    protected void notifyReload() {
        // Nothing to do
    }

    /**
     * Show osd.
     * 
     * @param audioObject
     *            the audio object
     */
    public void showOSD(AudioObject audioObject) {

        if (audioObject == null || VisualHandler.getInstance().getFullScreenWindow().isVisible()) {
            return;
        }

        // If the OSD is already visible stop animation
        stopAnimation(false);

        int x = 0;
        if (ApplicationState.getInstance().getOsdHorizontalAlignment() == SwingConstants.CENTER) {
            x = Toolkit.getDefaultToolkit().getScreenSize().width / 2 - ApplicationState.getInstance().getOsdWidth() / 2;
        } else if (ApplicationState.getInstance().getOsdHorizontalAlignment() == SwingConstants.LEFT) {
            x = 50;
        } else {
            x = Toolkit.getDefaultToolkit().getScreenSize().width - ApplicationState.getInstance().getOsdWidth() - 50;
        }
        x = Math.max(x, 0);

        int y = 0;
        if (ApplicationState.getInstance().getOsdVerticalAlignment() == SwingConstants.CENTER) {
            y = Toolkit.getDefaultToolkit().getScreenSize().height / 2 - getWindowControlled().getHeight() / 2;
        } else if (ApplicationState.getInstance().getOsdVerticalAlignment() == SwingConstants.TOP) {
            y = 20;
        } else {
            y = Toolkit.getDefaultToolkit().getScreenSize().height - 20 - getWindowControlled().getHeight();
        }

        locationToStay = new Point(x, y);
        hiddenLocation = new Point(locationToStay.x, Toolkit.getDefaultToolkit().getScreenSize().height + 4);

        if (audioObject instanceof PodcastFeedEntry) {
            getWindowControlled().setImage(ImageLoader.RSS);
            getWindowControlled().setLine1(audioObject.getTitle());
            getWindowControlled().setLine2(audioObject.getUrl());
            getWindowControlled().setLine3(audioObject.getArtist());
        } else if (audioObject instanceof Radio) {
            getWindowControlled().setImage(ImageLoader.RADIO);
            getWindowControlled().setLine1(audioObject.getTitle());
            getWindowControlled().setLine2(audioObject.getUrl());
            getWindowControlled().setLine3("");
        } else if (audioObject instanceof AudioFile) {
            // For Audio Files first try to get image from audio file or folder
            ImageIcon image = AudioFilePictureUtils.getImageForAudioFile((AudioFile) audioObject, -1, -1);
            if (image == null) {
                Image imageFromContext = ContextHandler.getInstance().getAlbumCover((AudioFile) audioObject);
                if (imageFromContext != null) {
                    image = new ImageIcon(imageFromContext);
                }
            }

            getWindowControlled().setImage(image);
            getWindowControlled().setLine1(StringUtils.getString(audioObject.getTitleOrFileName(), " (", TimeUtils.seconds2String(audioObject.getDuration()), ")"));
            getWindowControlled().setLine2(audioObject.getAlbum());
            getWindowControlled().setLine3(audioObject.getArtist());
        } else {
            throw new IllegalArgumentException(audioObject + " is not a audio file, radio or podcast feed entry");
        }
        if (SystemProperties.isJava6Update10OrLater) {
            getWindowControlled().setLocation(locationToStay);
            GuiUtils.setWindowOpacity(getWindowControlled(), 0);
            getWindowControlled().setRoundedBorders(true);
        } else {
            getWindowControlled().setLocation(ApplicationState.getInstance().isAnimateOSD() ? hiddenLocation : locationToStay);
        }
        getWindowControlled().setVisible(true);
        // see bug 1864517
        getWindowControlled().repaint();
        timer = getTimer();
        timer.start();
    }
}
