/*
 * aTunes 1.13.3
 * Copyright (C) 2006-2009 Alex Aranda, Sylvain Gaudard, Thomas Beckers and contributors
 *
 * See http://www.atunes.org/wiki/index.php?title=Contributing for information about contributors
 *
 * http://www.atunes.org
 * http://sourceforge.net/projects/atunes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

package net.sourceforge.atunes.kernel.modules.navigator;

import java.awt.event.MouseEvent;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Map;

import javax.swing.ImageIcon;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeCellRenderer;
import javax.swing.tree.TreePath;

import net.sourceforge.atunes.gui.model.NavigationTableModel;
import net.sourceforge.atunes.kernel.ControllerProxy;
import net.sourceforge.atunes.kernel.controllers.navigation.NavigationController.ViewMode;
import net.sourceforge.atunes.kernel.modules.state.ApplicationState;
import net.sourceforge.atunes.misc.log.LogCategories;
import net.sourceforge.atunes.misc.log.Logger;
import net.sourceforge.atunes.model.AudioObject;

public abstract class NavigationView {

    /**
     * Common logger for all navigation views
     */
    private static Logger logger = new Logger();

    /**
     * Scroll pane used for tree
     */
    private JScrollPane scrollPane;

    /**
     * @return the title of this view
     */
    public abstract String getTitle();

    /**
     * @return the icon of this view
     */
    public abstract ImageIcon getIcon();

    /**
     * @return the tooltip of this view in tabbed pane
     */
    public abstract String getTooltip();

    /**
     * @return the JTree that contains this view
     */
    public abstract JTree getTree();

    /**
     * Return tree renderer of this view
     * 
     * @return
     */
    protected abstract TreeCellRenderer getTreeRenderer();

    /**
     * 
     * @return the tree popup menu of this view
     */
    public abstract JPopupMenu getTreePopupMenu();

    /**
     * 
     * @return the table popup menu of this view
     */
    public JPopupMenu getTablePopupMenu() {
        // By default table popup is the same of tree
        return getTreePopupMenu();
    }

    /**
     * Returns scroll pane of tree
     * 
     * @return
     */
    public final JScrollPane getTreeScrollPane() {
        if (scrollPane == null) {
            scrollPane = new JScrollPane(getTree());
        }
        return scrollPane;
    }

    /**
     * Returns the data to be shown in the view. It depends on the view mode
     * 
     * @param viewMode
     * @return
     */
    protected abstract Map<String, ?> getViewData(ViewMode viewMode);

    /**
     * Refreshes view
     * 
     * @param viewMode
     * @param treeFilter
     */
    public void refreshView(ViewMode viewMode, String treeFilter) {
        // Get selected rows before refresh
        List<AudioObject> selectedObjects = ((NavigationTableModel) getNavigationTable().getModel()).getAudioObjectsAt(getNavigationTable().getSelectedRows());

        // Call to refresh tree
        refreshTree(viewMode, treeFilter);

        // Set the same selected audio objects as before refreshing
        for (AudioObject audioObject : selectedObjects) {
            int indexOfAudioObject = ((NavigationTableModel) getNavigationTable().getModel()).getSongs().indexOf(audioObject);
            if (indexOfAudioObject != -1) {
                getNavigationTable().addRowSelectionInterval(indexOfAudioObject, indexOfAudioObject);
            }
        }
    }

    /**
     * Refresh tree view
     * 
     * @param viewMode
     * @param treeFilter
     */
    protected abstract void refreshTree(ViewMode viewMode, String treeFilter);

    /**
     * Returns a list of audio object associated to a tree node
     * 
     * @param node
     * @param viewMode
     * @param treeFilter
     * @return
     */
    public abstract List<AudioObject> getAudioObjectForTreeNode(DefaultMutableTreeNode node, ViewMode viewMode, String treeFilter);

    /**
     * Returns <code>true</code> is audio objects associated to a tree node need
     * to be sorted
     * 
     * @return
     */
    public abstract boolean isAudioObjectsFromNodeNeedSort();

    /**
     * Returns the class of a given table column in this view
     * 
     * @param columnIndex
     * @return
     */
    public abstract Class<?> getNavigatorTableColumnClass(int columnIndex);

    /**
     * Returns the number of table columns in this view
     * 
     * @return
     */
    public abstract int getNavigatorTableColumnCount();

    /**
     * Returns the name of a given table column in this view
     * 
     * @param columnIndex
     * @return
     */
    public abstract String getNavigatorTableColumnName(int columnIndex);

    /**
     * Returns the value of a table column given the audio object represented by
     * the row in this view
     * 
     * @param audioObject
     * @param columnIndex
     * @return
     */
    public abstract Object getNavigatorTableValueAt(AudioObject audioObject, int columnIndex);

    /**
     * Returns the maximum with of the given column in this view
     * 
     * @param columnIndex
     * @return
     */
    public abstract int getNavigatorTableMaxWidthForColumn(int columnIndex);

    /**
     * Returns <code>true</code> if the view supports organize information in
     * different view modes
     * 
     * @return
     */
    public abstract boolean isViewModeSupported();

    /**
     * Returns <code>true</code> if the view supports filter navigation table
     * 
     * @return
     */
    public abstract boolean isNavigatorTableFilterSupported();

    /**
     * Returns a filtered list of audio objects given the original list of audio
     * objects and the filter to apply
     * 
     * @param audioObjects
     * @param filter
     * @return
     */
    public abstract List<AudioObject> filterNavigatorTable(List<AudioObject> audioObjects, String filter);

    /**
     * Enables or disables tree popup menu items of this view
     * 
     * @param e
     */
    public abstract void updateTreePopupMenuWithTreeSelection(MouseEvent e);

    /**
     * Enables or disables table popup menu items of this view
     * 
     * @param table
     * @param e
     */
    public abstract void updateTablePopupMenuWithTableSelection(JTable table, MouseEvent e);

    /** Collator for locale sensitive string comparison */
    protected Collator collator = Collator.getInstance(ApplicationState.getInstance().getLocale().getLocale());

    /** The smart comparator ignores a leading "The" */
    protected Comparator<String> smartComparator = new Comparator<String>() {
        private String removeThe(String str) {
            if (str.toLowerCase().startsWith("the ") && str.length() > 4) {
                return str.substring(4);
            }
            return str;
        }

        @Override
        public int compare(String s1, String s2) {
            return collator.compare(removeThe(s1).toLowerCase(), removeThe(s2).toLowerCase());
        }
    };

    /** The comparator. */
    protected Comparator<String> comparator = new Comparator<String>() {
        @Override
        public int compare(String s1, String s2) {
            return collator.compare(s1.toLowerCase(), s2.toLowerCase());
        }
    };

    /**
     * Method to log debug messages
     * 
     * @param message
     */
    protected final void debug(String message) {
        logger.debug(LogCategories.VIEW, message);
    }

    /**
     * Give access to navigation table
     * 
     * @return
     */
    protected JTable getNavigationTable() {
        return ControllerProxy.getInstance().getNavigationController().getNavigationPanel().getNavigationTable();
    }

    /**
     * Returns the default comparator
     * 
     * @return the default comparator
     */
    public Comparator<String> getComparator() {
        return comparator;
    }
    
    /**
     * Return current view mode
     * @return
     */
    protected ViewMode getCurrentViewMode() {
        return ApplicationState.getInstance().getViewMode();
    }
    
    /**
     * Return current tree filter
     * @return
     */
    protected String getCurrentTreeFilter() {
        return ControllerProxy.getInstance().getNavigationController().getNavigationPanel().getTreeFilterPanel().getFilter(); 
    }

    /**
     * Return selected objects in this navigation view
     * @return
     */
    public List<AudioObject> getSelectedObjects() {
        List<AudioObject> selectedInTable = ((NavigationTableModel) getNavigationTable().getModel()).getAudioObjectsAt(getNavigationTable().getSelectedRows());
        if (selectedInTable.isEmpty()) {
            TreePath[] paths = getTree().getSelectionPaths();
            List<AudioObject> audioObjectsSelected = new ArrayList<AudioObject>();
            if (paths != null) {
                for (TreePath path : paths) {
                    audioObjectsSelected.addAll(getAudioObjectForTreeNode((DefaultMutableTreeNode) path.getLastPathComponent(), getCurrentViewMode(), getCurrentTreeFilter()));
                }
            }
            return audioObjectsSelected;
        }
        return selectedInTable;
    }
   
}
