/*
 * aTunes 1.13.4
 * Copyright (C) 2006-2009 Alex Aranda, Sylvain Gaudard, Thomas Beckers and contributors
 *
 * See http://www.atunes.org/wiki/index.php?title=Contributing for information about contributors
 *
 * http://www.atunes.org
 * http://sourceforge.net/projects/atunes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

package net.sourceforge.atunes.kernel.modules.player.mplayer;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;

import javax.swing.SwingUtilities;

import net.sourceforge.atunes.kernel.modules.player.PlayerHandler;
import net.sourceforge.atunes.kernel.modules.podcast.PodcastFeedEntry;
import net.sourceforge.atunes.kernel.modules.radio.Radio;
import net.sourceforge.atunes.kernel.modules.repository.audio.AudioFile;
import net.sourceforge.atunes.kernel.modules.repository.audio.CueTrack;
import net.sourceforge.atunes.misc.log.Logger;
import net.sourceforge.atunes.model.AudioObject;
import net.sourceforge.atunes.utils.ClosingUtils;

/**
 * The Class MPlayerOutputReader.
 */
abstract class MPlayerOutputReader extends Thread {

    /** The Constant logger. */
    protected static final Logger logger = new Logger();

    /** The handler. */
    protected MPlayerHandler handler;

    /** The in. */
    protected BufferedReader in;

    /** The submitted. */
    protected boolean submitted;

    /** The started. */
    protected boolean started;

    /** Is cue sheet. */
    protected static boolean isCueSheet;

    /** The length. */
    protected int length;

    /** The time. */
    protected int time;

    /** Cue sheet track start position */
    protected static int cueTrackStartPosition;

    /**
     * Instantiates a new m player output reader.
     * 
     * @param handler
     *            the handler
     * @param process
     *            the process
     */
    protected MPlayerOutputReader(MPlayerHandler handler, Process process) {
        this.handler = handler;
        this.in = new BufferedReader(new InputStreamReader(process.getInputStream()));
    }

    /**
     * New instance.
     * 
     * @param handler
     *            the handler
     * @param process
     *            the process
     * @param ao
     *            the ao
     * 
     * @return the mplayer output reader
     */
    static MPlayerOutputReader newInstance(MPlayerHandler handler, Process process, AudioObject ao) {
        if (ao instanceof AudioFile) {
            // In case of a cue sheet, correct the position to reflect the position in the cue track and not the one in the actual audio file.
            if (AudioFile.isCueFile(((AudioFile) ao).getFile())) {
                cueTrackStartPosition = ((CueTrack) ao).getTrackStartPositionAsInt() * 1000;
                isCueSheet = true;
            } else {
                isCueSheet = false;
            }
            return new AudioFileMPlayerOutputReader(handler, process, (AudioFile) ao);
        } else if (ao instanceof Radio) {
            return new RadioMPlayerOutputReader(handler, process, (Radio) ao);
        } else if (ao instanceof PodcastFeedEntry) {
            return new PodcastFeedEntryMPlayerOutputReader(handler, process, (PodcastFeedEntry) ao);
        } else {
            throw new IllegalArgumentException("audio object is not from type AudioFile, Radio or PodcastFeedEntry");
        }
    }

    /**
     * Inits.
     */
    protected void init() {
        // Nothing to do
    }

    /**
     * Read.
     * 
     * @param line
     *            the line
     */
    protected void read(String line) {
        // Read progress			
        // MPlayer bug: Duration still inaccurate with mp3 VBR files! Flac duration bug
        if (line.matches(".*ANS_TIME_POSITION.*")) {
            if (isCueSheet) {
                time = (int) (Float.parseFloat(line.substring(line.indexOf("=") + 1)) * 1000.0) - cueTrackStartPosition;
                handler.setTime(time);
            } else {
                time = (int) (Float.parseFloat(line.substring(line.indexOf("=") + 1)) * 1000.0);
                handler.setTime(time);
            }
        }

        // End
        if (line.matches(".*\\x2e\\x2e\\x2e\\x20\\(.*\\x20.*\\).*")) {
            handler.playNextAudioObject(true);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Thread#run()
     */
    @Override
    public final void run() {
        String line = null;
        try {
            SwingUtilities.invokeLater(new Runnable() {
                @Override
                public void run() {
                    init();
                }
            });
            while ((line = in.readLine()) != null && !isInterrupted()) {
                final String lineHelp = line;
                SwingUtilities.invokeLater(new Runnable() {
                    @Override
                    public void run() {
                        read(lineHelp);
                    }
                });
            }
        } catch (final IOException e) {
            SwingUtilities.invokeLater(new Runnable() {
                @Override
                public void run() {
                    PlayerHandler.notifyPlayerError(e);
                }
            });
        } finally {
            ClosingUtils.close(in);
        }
    }

}
